#include <gtk/gtk.h>
#include <stdio.h>
#include "tomoe.h"

/* ΰΤΥХå󥰥ԥåޥå */
static GdkPixmap *pixmap  = NULL;
static gint prev_x = -1;
static gint prev_y = -1;

static GList *stroke      = NULL;
static GList *stroke_list = NULL;

/* ŬڤʥοХå󥰥ԥåޥåפ */
static gint
configure_event (GtkWidget *widget, GdkEventConfigure *event)
{
  if (pixmap)
    gdk_pixmap_unref(pixmap);

  pixmap = gdk_pixmap_new(widget->window,
			  widget->allocation.width,
			  widget->allocation.height,
			  -1);
  gdk_draw_rectangle (pixmap,
		      widget->style->white_gc,
		      TRUE,
		      0, 0,
		      widget->allocation.width,
		      widget->allocation.height);

  return TRUE;
}

/* Хå󥰥ԥåޥåפ饹꡼ */
static gint
expose_event (GtkWidget *widget, GdkEventExpose *event)
{
    gdk_draw_pixmap(widget->window,
		    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
		    pixmap,
		    event->area.x, event->area.y,
		    event->area.x, event->area.y,
		    event->area.width, event->area.height);

    return FALSE;
}

static void
draw_brush (GtkWidget *widget, gint x, gint y)
{
    GdkRectangle update_rect;

    point *p = g_malloc(sizeof (point));
    p->x = x;
    p->y = y;
    stroke = g_list_append(stroke, p);

    if (prev_x == -1) {
	prev_x  = x;
	prev_y  = y;
    }
  
    update_rect.x = MIN(x, prev_x);
    update_rect.y = MIN(y, prev_y);
    update_rect.width  = abs(x - prev_x) + 1;
    update_rect.height = abs(y - prev_y) + 1;

    gdk_draw_line(pixmap, widget->style->black_gc, x, y, prev_x, prev_y);
    gtk_widget_draw(widget, &update_rect);

    prev_x = x;
    prev_y = y;
}

static gint
button_press_event (GtkWidget *widget, GdkEventButton *event)
{
    if (event->button == 1 && pixmap != NULL) {
	draw_brush(widget, event->x, event->y);
    }
    return TRUE;
}

static gint
button_release_event (GtkWidget *widget, GdkEventButton *event)
{
    stroke_list = g_list_append(stroke_list, stroke);
    stroke = NULL;
    prev_x = prev_y = -1;
    return TRUE;
}

static gint
motion_notify_event (GtkWidget *widget, GdkEventMotion *event)
{
    int x, y;
    GdkModifierType state;

    if (event->is_hint) {
	gdk_window_get_pointer(event->window, &x, &y, &state);
    } else {
	x = event->x;
	y = event->y;
	state = event->state;
    }
    
    if (state & GDK_BUTTON1_MASK && pixmap != NULL) {
	draw_brush (widget, x, y);
    }
    return TRUE;
}

static gint
get_distance (GList *first_node, GList *last_node, GList **most_node)
{
    /* Getting distance 
     * MAX( |aw - bv + c| )
     * a = x-p : b = y-q : c = py - qx
     * first = (p, q) : last = (x, y) : other = (v, w)
     */

    GList *dot;
    gint a, b, c;
    gint dist = 0;
    gint max  = 0;
    point *first = first_node->data;
    point *last  = last_node->data;
    point *p;

    *most_node = NULL;
    if (first_node == last_node) {
	return 0;
    }

    a = last->x - first->x;
    b = last->y - first->y;
    c = last->y * first->x - last->x * first->y;

    for (dot = first_node; dot != last_node; dot = dot->next) {
	p = dot->data;
	dist = abs((a * p->y) - (b * p->x) + c);
	if (dist > max) {
	    max = dist;
	    *most_node = dot;
	}
    }
    return max * max / (a*a + b*b);
}

static GList*
get_vertex (GList *first_node, GList *last_node)
{
    GList *rv = NULL;
    GList *most_node;
    gint dist;
    gint error = 300 * 300 / 400; /* 5% */

    dist = get_distance(first_node, last_node, &most_node);

    if (dist > error) {
	rv = g_list_concat(get_vertex(first_node, most_node),
			   get_vertex(most_node, last_node));
    } else {
	rv = g_list_append(rv, last_node->data);
    }
    return rv;
}

static void
clear (GtkWidget *widget)
{
    GdkRectangle update_rect;
    GList *line, *dots, *dot;
    gint x, y, px, py;

    update_rect.x = 0;
    update_rect.y = 0;
    update_rect.width  = widget->allocation.width;
    update_rect.height = widget->allocation.height;

    gdk_draw_rectangle(pixmap, widget->style->white_gc, TRUE, 0, 0,
		       widget->allocation.width,
		       widget->allocation.height);
    
    if(g_list_length(stroke_list) == 0) {
	stroke_list = NULL;
	gtk_widget_draw (widget, &update_rect);
	return;
    }

    g_print(":%d\n", g_list_length(stroke_list));
    for (line = stroke_list; line; line = line->next) {
	dots = g_list_prepend(get_vertex(line->data, g_list_last(line->data)),
			      ((GList *)(line->data))->data);
	px = py = -1;
	g_print("%d ", g_list_length(dots));
	for (dot = dots; dot; dot = dot->next) {
	    x = ((point *)(dot->data))->x;
	    y = ((point *)(dot->data))->y;
	    g_print("(%d %d) ", x, y);
	    if (px != -1) {
		gdk_draw_line(pixmap, widget->style->black_gc, px, py, x, y);
	    }
	    px = x;
	    py = y;
	}
	g_print("\n");
    }
    g_print("\n");
    
    stroke_list = NULL;
    gtk_widget_draw (widget, &update_rect);
}    

void
quit ()
{
    g_print("[EOF]\n");
    gtk_exit (0);
}

int
main (int argc, char *argv[])
{
  GtkWidget *window;
  GtkWidget *drawing_area;
  GtkWidget *vbox;

  GtkWidget *button;

  gtk_init (&argc, &argv);

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_name (window, "Test Input");

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show (vbox);

  gtk_signal_connect (GTK_OBJECT (window), "destroy",
		      GTK_SIGNAL_FUNC (quit), NULL);

  /* ΰ  */

  drawing_area = gtk_drawing_area_new ();
  gtk_drawing_area_size (GTK_DRAWING_AREA (drawing_area), 300, 300);
  gtk_box_pack_start (GTK_BOX (vbox), drawing_area, TRUE, TRUE, 0);

  gtk_widget_show (drawing_area);

  /* Хå󥰥ԥåޥåפν˻Ѥ륷ʥ */

  gtk_signal_connect (GTK_OBJECT (drawing_area), "expose_event",
		      (GtkSignalFunc) expose_event, NULL);
  gtk_signal_connect (GTK_OBJECT(drawing_area),"configure_event",
		      (GtkSignalFunc) configure_event, NULL);

  /* ٥ȥʥ */

  gtk_signal_connect (GTK_OBJECT (drawing_area), "motion_notify_event",
		      (GtkSignalFunc) motion_notify_event, NULL);
  gtk_signal_connect (GTK_OBJECT (drawing_area), "button_press_event",
		      (GtkSignalFunc) button_press_event, NULL);
  gtk_signal_connect (GTK_OBJECT (drawing_area), "button_release_event",
		      (GtkSignalFunc) button_release_event, NULL);

  gtk_widget_set_events (drawing_area, GDK_EXPOSURE_MASK
			 | GDK_LEAVE_NOTIFY_MASK
			 | GDK_BUTTON_PRESS_MASK
			 | GDK_BUTTON_RELEASE_MASK
			 | GDK_POINTER_MOTION_MASK
			 | GDK_POINTER_MOTION_HINT_MASK);

  button = gtk_button_new_with_label ("Clear");
  gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 0);

  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     GTK_SIGNAL_FUNC (clear),
			     GTK_OBJECT (drawing_area));
  gtk_widget_show (button);

  gtk_widget_show (window);

  gtk_main ();

  return 0;
}
/* example-end */
